#!/usr/bin/perl -w
use strict;
my $rfbdfile = $ARGV[0];


# this is a simple download manager for RFBD audioplus downloads.
#
# It reads the rfbd file that is provided when you download a book.
# This file provides information about which .zip files need to be
# downloaded, how big they are, and how many.  It is provides the title
# of the book.  All this informaiton is used to download the .zip files
# and build an appropriate folder structure for storing the .ZiP files.

# Path to the top level directory where the script should put the
# downloaded  zip files.  Each book/CD will have a subfolder.

#my $DL_PREFIX = $ENV{"DOWNLOADS"} || die "Please set the DOWNLOADS envvironment variable.\n";
my $DL_PREFIX = "$ARGV[1]";
# Where your browser saves the *.rfbd files that are downloaded fromt he
# "my downloads" page.
#my $DL_DIR = $ENV{"RFBD_DIRECTORY"} || die "Please set the RFBD_DIRECTORY envvironment variable.\n";
#my $DL_DIR = "/Users/gkearney/RFBD/RFBD_DIRECTORY";

use LWP::Simple;
use File::Copy;

# Global variable that is set to 1 if the -q flag is on the command line.
# HACKY!
my $flag_quiet = 0;

# list of the zipped files downloaded for a particular book.
my @zip_files = ();

# the .rfbd file contains records, each describing one zip file.  The
# file parser creates a table of key/value pairs and calls this function
# for each zip file.
#
# TODO: consider forking the download into its own process so we can
# parallelize the downloads.  Easy enough to manage this, and limiting
# to a fix number of parallel downloads.

sub process_record
{
    my $record = shift;

    # Validate that the table is good.  We only have correct keys, and
    # their values pass simple syntax checks.
    foreach my $key (keys %$record)
    {
	if ($key eq "URL")
	{
	    die $record->{$key} . ": Invalid URL.\n"
	         unless $record->{$key} =~ m!^http://.*/.*\.zip$!i;
	}
	elsif ($key eq "Desc")
	{
	     die $record->{$key} . ": Invalid description.\n"
	         unless $record->{$key} =~ m!^(.*?)\s*\(.*?\):\s*File\s+(\d+)\s+of\s+(\d+)$!;

	     # Replace the existing "Desc" value with just the book title.
	     $record->{$key} = $1;

	     # Store the broken out count and total numbers in their own
	     # table entries.  Keys that begin with _ are assumed to be 
	     # generated by this function and are not sanity checked.
	     $record->{"_count"} = $2;
	     $record->{"_total"} = $3;
	}
	elsif ($key eq "Folder")
	{
	     # Compalin if we find any character which is not a valid VFAT 
	     # filename character.  Stole this set from wikipedia
	     die $record->{$key} . ": Invalid folder.\n"
	         if $record->{$key} =~ /[^-a-zA-Z0-9!#$%&'"()@^_`{}~\/\\]/;
	}
	elsif ($key eq "Size")
	{
	     die $record->{$key} . ": Invalid size.\n"
	         unless $record->{$key} =~ /^\d+$/;
	}
	# The value of WMP is"No" unless this is a windows media
	# download, which we don't deal with.
	elsif ($key eq "WMP")
	{
	     die $record->{$key} . ": does not say this is not a WMP record.\n"
	         unless lc $record->{$key} eq "no";
	}
	# locally generated keys start with _ and are ignored by this
	# validator.
	elsif (substr($key,0,1) ne "_")
	{
	     die "$key: is an invalid key.  Value = " . $record->{$key} . "\n";
	}
    }

    # fix the folder.  Strip off training \, change the shelf number to
    # the book title, and change \ to / throughout.
    $record->{"Folder"} =~ s!\\$!!;
    $record->{"Folder"} =~ s!^.*?\\!$record->{"Desc"}\\!g;
    $record->{"Folder"} =~ s!\\!/!g;

    # Prepend the top level directory where we are going to store the
    # zip files to the folder name.
    $record->{"Folder"} = $DL_PREFIX . "/" . $record->{"Folder"};

    # Get the zipfile name by stripping everything off the URL upto the
    # last / character.
    $record->{"Name"} = $record->{"URL"};
    $record->{"Name"} =~ s!^.*/!!;

    # make sure directory (folder) we're going to store the zip files in
    # exists.
    mkdirhier($record->{"Folder"});

    my $file = $record->{"Folder"} . "/" . $record->{"Name"};
    my $tag = sprintf("%3d / %3d: %s",$record->{"_count"},$record->{"_total"},$record->{"Desc"});

    # if the zip file already exists, and the size matches what's int he
    # config, skip it.
    if (-e $file && -s _ eq $record->{"Size"})
    {
	print "Already downloaded " .  $tag . "\n" unless $flag_quiet;
	push (@zip_files,[$record->{"Folder"}, $file]);
    }
    # otherwise, download it using LWP::Simple::getstore
    else
    {
	print "Downloading " . $tag . "\n" unless $flag_quiet;
	my $status = getstore($record->{"URL"}, $file);
	if ($status != 200)
	{
	    unlink($file);
	    die $status, ": failed to download ", $record->{"URL"}, "\n";
	}
	else
	{
	    push (@zip_files,[$record->{"Folder"}, $file]);
	}
    }
}

# make every missing directory in path.  i.e. mkdirhier /x/y/z will
# create directories /x /x/y and /x/y/z if they don't exist.
sub mkdirhier
{
    my $folder = shift;
    # turn // into / and strip trailing /
    $folder =~ s!//+!/!g;
    $folder =~ s!/+$!!g;
    my @dirs = split(m!/!, $folder);

    my $build = "";
    for my $dir (@dirs)
    {
        if ($dir eq "")
	{
	    $build .= "/";
	}
	else
	{
	    $build .= $dir . "/";
	}

	unless (-d $build)
	{
	    mkdir $build, 0777 or die "mkdir $build: $!\n";
	}
    }
}

my @paths = split(/:/,$ENV{"PATH"});

my $found_unzip = 0;
for my $path (@paths)
{
    if (-x $path . "/" . "unzip")
    {
        $found_unzip = 1;
	last;
    }
}

die "unzip is not in your path\n" unless $found_unzip;

# TODO: put in a real flag processing engine, or use someone else's.
#if (@ARGV > 0 && $ARGV[0] eq "-q")
#{
#    shift;
#    $flag_quiet = 1;
#}

#die "usage: download [-q]\n" if @ARGV != 0;

# Table that holds the key/value pairs for each zip file defined in
# .rfbd files.
my %table = ();

# Track if we have found part of a record and are working to build the
# rest of it.
my $in_record = 0;

# look for *.rfbd files in ${DL_DIR}
#my @inputs = <${DL_DIR}/*.rfbd>;

my @inputs = $rfbdfile;
die "No *.rfbd files in /tmp\n" if @inputs < 1;
for my $input (@inputs)
{
    open(INPUT,"< $input") or die "$input: $!\n";

    # Parse the .rfbd config
    while (<INPUT>)
    {
	# Strip leading/trailing blanks.
	s!^\s*(.*?)\s*$!$1!;
	# Skip comment lines, lines that *begin* with //.  Note that it
	# is not true that comments start anywhere there's a //
	next if m!^//!;

	# If we found a blank line, that means the end of a record.
	if ($_ eq "")
	{
	    # but only if we had found the beginning of a record already.
	    if ($in_record)
	    {
		process_record(\%table);
		$in_record = 0;
		%table = ();
	    }
	}
	else
	{
	    # Split the key: value pair on the first colon only.
	    my @elems = split(/\s*:\s*/,$_,2);
	    # Add to the table.
	    $table{$elems[0]} = $elems[1];
	    # make it clear we're building a record.
	    $in_record = 1;
	}
    }

    # end of the .rfbd file reached.  If we were building a record, process it.
    process_record(\%table) if $in_record;

    close (INPUT);

    # Now, unzip the files.  Do this at the end, so that we can
    # correctly detect an interrupted download.

    print "Unzipping files\n" unless $flag_quiet;

    map { 
        print "unzip $_->[1]\n"
	    unless $flag_quiet;
        system          "unzip", "-o", "-L", "-q", "-d",$_->[0], $_->[1];
	die "Unzip problem with $_[1]\n" unless ($? >> 8) == 0;
    } @zip_files;

    # PUT the .rfbd file in the top level directory for safe keeping.
    # TODO: consider putting them in a .archive directory.
    die "move($input, $DL_PREFIX): $!\n" 
         unless move($input, $DL_PREFIX);

    # Delete the zip files.
    map { 
        print "Deleting " . $_->[1] . "\n" unless $flag_quiet;
	unlink($_->[1]) or die "Unlink $_->[1]: $!\n";
    } @zip_files;
    
    if (-e "/Library/PreferecePanes/Growl.prefPane") {
    
    `./growlnotify -m "Finished Downloading from RFB&D"`;
    }
}
